/* 
DeGrainMedian plugin for Avisynth 2.5 - Spatio-Temporal Limited Median filter
Version 0.8 
Copyright (C)2004-2005 Alexander G. Balakhnin aka Fizick.
bag@hotmail.ru         http://bag.hotmail.ru
under the GNU General Public Licence version 2.

 The plugin is based on STMedianFilter plugin by Tom Barry
 trbarry@trbarry.com,
 Some ASM macros borrowed from RemoveGrain code by Rainer Wittmann <gorw@gmx.de> aka Kassandro


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Change Log

2005/10/08 v0.8 - improved cache (range=3);
                - added option norow to disable same row (horizontal) spatial filtering;
				- renamed old mode 2 to 3, 3 to 4, 4 to 5, added new mode 2, 
				- fixed documentation error with mode 1
2005/04/21 v 0.7 - fixed bug with mode=2 for YV12 interlaced; first frame pitch fixed  
2007/04/02  v 0.6 fixed some bugs with YUY2 (thanks to AVIL for report)
2005/03/23  v.0.5    Use more correct cache settings. Thank to Turyst04 for stablity problem report.
2004/09/10  v.0.4   Fixed possible bug with different frame pitches
2004/08/28  V 0.3	Added interlaced mode, code reorganization
2004/08/21  V 0.2.0.0	more safe mode=2 is added 
2004/08/17  V 0.1.1.0	Bug fixed for spatial part of YV12 mode 
2004/08/13  V 0.1.0.0	Initial release 

*/

#include "windows.h"
#include "avisynth.h"

class DeGrainMedian : public GenericVideoFilter {
    int  Max_Frames;
    int  limitY;
    int  limitUV;
	int	 mode;
	bool interlaced;
	bool norow;

	int  useFrame;
	bool SSEMMXenabled;
	bool DoYV12;

public:
    DeGrainMedian(PClip _child, int _limitY, int _limitUV, int _mode, bool _interlaced, bool _norow, IScriptEnvironment* env) :
	  GenericVideoFilter(_child), limitY(_limitY), limitUV(_limitUV), mode(_mode), interlaced(_interlaced), norow(_norow)
    {
        Max_Frames = vi.num_frames;
		SSEMMXenabled = (env->GetCPUFlags() & CPUF_INTEGER_SSE) !=0;


	    if (vi.IsYUY2())
			DoYV12 = false;
		else if (vi.IsYV12())
			DoYV12 = true;
	    else
			env->ThrowError("DeGrainMedian: Supports YUY2 and YV12 color formats only");

	    if (!SSEMMXenabled)
			env->ThrowError("DeGrainMedian: Only SSEMMX CPU's currently supported");

		if ( limitY<0 || limitY>255 || limitUV<0 || limitUV>255 )
			env->ThrowError("DeGrainMedian: Limit must be from 0 to 255");

		if ( mode<0 || mode>5 )
			env->ThrowError("DeGrainMedian: Mode must be from 0 to 5"); // v0.8

		_child->SetCacheHints(CACHE_RANGE,3); // added (2) in v.0.5, replaced to 3 in v.0.8

    }

    PVideoFrame __stdcall GetFrame(int inFrame, IScriptEnvironment* env);

};

//------------------------------------------------------------------------
// declarations
	void DeGrainMedian_YUY2mode0(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode1(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode2(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode3(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode4(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode5(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				

	void DeGrainMedian_YUY2mode0i(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode1i(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode2i(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode3i(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode4i(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode5i(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				

	void DeGrainMedian_YV12mode0(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode1(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode2(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode3(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode4(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode5(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			

	void DeGrainMedian_YV12mode0i(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode1i(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode2i(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode3i(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode4i(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode5i(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			

	void DeGrainMedian_YUY2mode0norow(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode1norow(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode2norow(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode3norow(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode4norow(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode5norow(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				

	void DeGrainMedian_YUY2mode0norowi(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode1norowi(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode2norowi(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode3norowi(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode4norowi(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				
	void DeGrainMedian_YUY2mode5norowi(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext,
		int rowsize, int FldHeight, int LimitY, int LimitUV);				

	void DeGrainMedian_YV12mode0norow(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode1norow(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode2norow(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode3norow(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode4norow(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode5norow(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			

	void DeGrainMedian_YV12mode0norowi(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode1norowi(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode2norowi(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode3norowi(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode4norowi(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			
	void DeGrainMedian_YV12mode5norowi(BYTE* dstp, int dst_pitch, const BYTE* pPrev, 
		const BYTE* srcp, int src_pitch,  const BYTE* pNext, 
		int rowsize, int FldHeight, int limit);			

//---------------------------------------------------------------------------------

PVideoFrame __stdcall DeGrainMedian::GetFrame(int inFrame, IScriptEnvironment* env) 
{
PVideoFrame PrevSrc;
PVideoFrame NextSrc;
PVideoFrame dst;

const BYTE* srcp;
const BYTE* pPrev;
const BYTE* pNext;
 BYTE * prevData;
 BYTE * nextData;
 BYTE * dstp;
bool prev_created = false;
bool next_created = false;
int src_pitch, prev_pitch, next_pitch, dst_pitch, rowsize, height;
	
	useFrame = inFrame;
	
	// Most of our processing is the same for YUY2 or YV12.
	PVideoFrame src = child->GetFrame(useFrame, env);

//	if (useFrame < 1 || useFrame >= Max_Frames-1)
//	{
//		return src;						// don't filter first and last frames
//	}

	dst = env->NewVideoFrame(vi);

	if (useFrame >= 1 && useFrame < Max_Frames-1)
	{
		PrevSrc = child->GetFrame(useFrame-1, env);
		NextSrc = child->GetFrame(useFrame+1, env);
	}


	if (interlaced) 
	{
		// But if it is YV12 we still have to copy the 2 chroma planes
		if (vi.IsYV12())
		{
			srcp = src->GetReadPtr(PLANAR_Y); // moved here in v 0.7
			dstp = dst->GetWritePtr(PLANAR_Y);
			src_pitch = src->GetPitch(PLANAR_Y);
			dst_pitch = dst->GetPitch(PLANAR_Y);
			rowsize = dst->GetRowSize(PLANAR_Y);
			height = dst->GetHeight(PLANAR_Y);
			if (limitY>0 && useFrame >= 1 && useFrame < Max_Frames-1)
			{
				pPrev = PrevSrc->GetReadPtr(PLANAR_Y);
				pNext = NextSrc->GetReadPtr(PLANAR_Y);

				prev_pitch = PrevSrc->GetPitch(PLANAR_Y);
				next_pitch = NextSrc->GetPitch(PLANAR_Y);
				// check pitches and make its equal by copying (for algo) - added in v.0.4
				prev_created = false;
				if (prev_pitch != src_pitch)
				{ // copy frame data to array with same pitch as source 
					prevData = (BYTE *)malloc(height*src_pitch);
					prev_created = true;
					env->BitBlt(prevData, src_pitch, pPrev, prev_pitch, rowsize, height);
					pPrev = prevData;
				}
				next_created = false;
				if (next_pitch != src_pitch)
				{ // copy frame data to array with same pitch as source (for algo)
					nextData = (BYTE *)malloc(height*src_pitch);
					next_created = true;
					env->BitBlt(nextData, src_pitch, pNext, next_pitch, rowsize, height);
					pNext = nextData;
				}

				if (norow)
				{
					if (mode==0)
						DeGrainMedian_YV12mode0norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==1)
						DeGrainMedian_YV12mode1norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==2)
						DeGrainMedian_YV12mode2norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==3)
						DeGrainMedian_YV12mode3norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==4)
						DeGrainMedian_YV12mode4norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==5)
						DeGrainMedian_YV12mode5norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
				}
				else
				{
					if (mode==0)
						DeGrainMedian_YV12mode0i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==1)
						DeGrainMedian_YV12mode1i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==2)
						DeGrainMedian_YV12mode2i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==3)
						DeGrainMedian_YV12mode3i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==4)
						DeGrainMedian_YV12mode4i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==5)
						DeGrainMedian_YV12mode5i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
				}


				// free created copy arrays
				if (prev_created == true) 
					free(prevData);
				if (next_created == true) 
					free(nextData);
			}
			else
				env->BitBlt(dstp, dst_pitch, srcp, src_pitch, rowsize, height);
			
			srcp = src->GetReadPtr(PLANAR_U);
			dstp = dst->GetWritePtr(PLANAR_U);
			src_pitch = src->GetPitch(PLANAR_U);
			dst_pitch = dst->GetPitch(PLANAR_U);
			rowsize = dst->GetRowSize(PLANAR_U);
			height = dst->GetHeight(PLANAR_U);
			if (limitUV>0 && useFrame >= 1 && useFrame < Max_Frames-1)
			{
				pPrev = PrevSrc->GetReadPtr(PLANAR_U);
				pNext = NextSrc->GetReadPtr(PLANAR_U);
				prev_pitch = PrevSrc->GetPitch(PLANAR_U);
				next_pitch = NextSrc->GetPitch(PLANAR_U);
				// check pitches and make its equal by copying (for algo) - added in v.0.4
				prev_created = false;
				if (prev_pitch != src_pitch)
				{ // copy frame data to array with same pitch as source 
					prevData = (BYTE *)malloc(height*src_pitch);
					prev_created = true;
					env->BitBlt(prevData, src_pitch, pPrev, prev_pitch, rowsize, height);
					pPrev = prevData;
				}
				next_created = false;
				if (next_pitch != src_pitch)
				{ // copy frame data to array with same pitch as source (for algo)
					nextData = (BYTE *)malloc(height*src_pitch);
					next_created = true;
					env->BitBlt(nextData, src_pitch, pNext, next_pitch, rowsize, height);
					pNext = nextData;
				}

				if (norow)
				{
					if (mode==0)
						DeGrainMedian_YV12mode0norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext,rowsize, height, limitUV);
					else if (mode==1)
						DeGrainMedian_YV12mode1norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==2)
						DeGrainMedian_YV12mode2norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==3)
						DeGrainMedian_YV12mode3norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==4)
						DeGrainMedian_YV12mode4norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==5)
						DeGrainMedian_YV12mode5norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
				}
				else
				{
					if (mode==0)
						DeGrainMedian_YV12mode0i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext,rowsize, height, limitUV);
					else if (mode==1)
						DeGrainMedian_YV12mode1i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==2)
						DeGrainMedian_YV12mode2i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==3)
						DeGrainMedian_YV12mode3i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==4)
						DeGrainMedian_YV12mode4i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==5)
						DeGrainMedian_YV12mode5i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
				}
			

				// free created copy arrays
				if (prev_created == true) 
					free(prevData);
				if (next_created == true) 
					free(nextData);
			}
			else
				env->BitBlt(dstp, dst_pitch, srcp, src_pitch, rowsize, height);
			
			srcp = src->GetReadPtr(PLANAR_V);
			dstp = dst->GetWritePtr(PLANAR_V);
			src_pitch = src->GetPitch(PLANAR_V);
			dst_pitch = dst->GetPitch(PLANAR_V);
			rowsize = dst->GetRowSize(PLANAR_V);
			height = dst->GetHeight(PLANAR_V);
			if (limitUV>0 && useFrame >= 1 && useFrame < Max_Frames-1)
			{
				pPrev = PrevSrc->GetReadPtr(PLANAR_V);
				pNext = NextSrc->GetReadPtr(PLANAR_V);
				prev_pitch = PrevSrc->GetPitch(PLANAR_V);
				next_pitch = NextSrc->GetPitch(PLANAR_V);
				// check pitches and make its equal by copying (for algo) - added in v.0.4
				prev_created = false;
				if (prev_pitch != src_pitch)
				{ // copy frame data to array with same pitch as source 
					prevData = (BYTE *)malloc(height*src_pitch);
					prev_created = true;
					env->BitBlt(prevData, src_pitch, pPrev, prev_pitch, rowsize, height);
					pPrev = prevData;
				}
				next_created = false;
				if (next_pitch != src_pitch)
				{ // copy frame data to array with same pitch as source (for algo)
					nextData = (BYTE *)malloc(height*src_pitch);
					next_created = true;
					env->BitBlt(nextData, src_pitch, pNext, next_pitch, rowsize, height);
					pNext = nextData;
				}

				if (norow)
				{
					if (mode==0)
						DeGrainMedian_YV12mode0norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext,rowsize, height, limitUV);
					else if (mode==1)
						DeGrainMedian_YV12mode1norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==2)
						DeGrainMedian_YV12mode2norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==3)
						DeGrainMedian_YV12mode3norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==4)
						DeGrainMedian_YV12mode4norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==5)
						DeGrainMedian_YV12mode5norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
				}
				else
				{
					if (mode==0)
						DeGrainMedian_YV12mode0i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext,rowsize, height, limitUV);
					else if (mode==1)
						DeGrainMedian_YV12mode1i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==2)
						DeGrainMedian_YV12mode2i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==3)
						DeGrainMedian_YV12mode3i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==4)
						DeGrainMedian_YV12mode4i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==5)
						DeGrainMedian_YV12mode5i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
				}
			

				// free created copy arrays
				if (prev_created == true) 
					free(prevData);
				if (next_created == true) 
					free(nextData);
			}
			else
				env->BitBlt(dstp, dst_pitch, srcp, src_pitch, rowsize, height);
			
		}
		else // YUY2
		{
			srcp = src->GetReadPtr();
			dstp = dst->GetWritePtr();
			src_pitch = src->GetPitch();
			dst_pitch = dst->GetPitch();
			rowsize = dst->GetRowSize();
			height = dst->GetHeight();
			if(useFrame >= 1 && useFrame < Max_Frames-1)
			{
				pPrev = PrevSrc->GetReadPtr();
				pNext = NextSrc->GetReadPtr();
				prev_pitch = PrevSrc->GetPitch();
				next_pitch = NextSrc->GetPitch();
				// check pitches and make its equal by copying (for algo) - added in v.0.4
				prev_created = false;
				if (prev_pitch != src_pitch)
				{ // copy frame data to array with same pitch as source 
					prevData = (BYTE *)malloc(height*src_pitch);
					prev_created = true;
					env->BitBlt(prevData, src_pitch, pPrev, prev_pitch, rowsize, height);
					pPrev = prevData;
				}
				next_created = false;
				if (next_pitch != src_pitch)
				{ // copy frame data to array with same pitch as source (for algo)
					nextData = (BYTE *)malloc(height*src_pitch);
					next_created = true;
					env->BitBlt(nextData, src_pitch, pNext, next_pitch, rowsize, height);
					pNext = nextData;
				}

				if (norow)
				{
					if (mode==0)
						DeGrainMedian_YUY2mode0norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY, limitUV);	
					else if (mode==1)
						DeGrainMedian_YUY2mode1norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY, limitUV);	
					else if (mode==2)
						DeGrainMedian_YUY2mode2norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY, limitUV);	
					else if (mode==3)
						DeGrainMedian_YUY2mode3norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY, limitUV);	
					else if (mode==4)
						DeGrainMedian_YUY2mode4norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY, limitUV);	
					else if (mode==5)
						DeGrainMedian_YUY2mode5norowi( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY, limitUV);	
				}
				else
				{
					if (mode==0)
						DeGrainMedian_YUY2mode0i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY, limitUV);	
					else if (mode==1)
						DeGrainMedian_YUY2mode1i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY, limitUV);	
					else if (mode==2)
						DeGrainMedian_YUY2mode2i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY, limitUV);	
					else if (mode==3)
						DeGrainMedian_YUY2mode3i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY, limitUV);	
					else if (mode==4)
						DeGrainMedian_YUY2mode4i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY, limitUV);	
					else if (mode==5)
						DeGrainMedian_YUY2mode5i( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY, limitUV);	
				}

				// free created copy arrays
				if (prev_created == true) 
					free(prevData);
				if (next_created == true) 
					free(nextData);
			}
			else
				env->BitBlt(dstp, dst_pitch, srcp, src_pitch, rowsize, height);
		}

	}
	else
	{ // progressive
		// But if it is YV12 we still have to copy the 2 chroma planes
		if (vi.IsYV12())
		{
			srcp = src->GetReadPtr(PLANAR_Y);
			dstp = dst->GetWritePtr(PLANAR_Y);
			src_pitch = src->GetPitch(PLANAR_Y);
			dst_pitch = dst->GetPitch(PLANAR_Y);
			rowsize = dst->GetRowSize(PLANAR_Y);
			height = dst->GetHeight(PLANAR_Y);
			if (limitY>0 && useFrame >= 1 && useFrame < Max_Frames-1)
			{
				pPrev = PrevSrc->GetReadPtr(PLANAR_Y);
				pNext = NextSrc->GetReadPtr(PLANAR_Y);
				prev_pitch = PrevSrc->GetPitch(PLANAR_Y);
				next_pitch = NextSrc->GetPitch(PLANAR_Y);
				// check pitches and make its equal by copying (for algo) - added in v.0.4
				prev_created = false;
				if (prev_pitch != src_pitch)
				{ // copy frame data to array with same pitch as source 
					prevData = (BYTE *)malloc(height*src_pitch);
					prev_created = true;
					env->BitBlt(prevData, src_pitch, pPrev, prev_pitch, rowsize, height);
					pPrev = prevData;
				}
				next_created = false;
				if (next_pitch != src_pitch)
				{ // copy frame data to array with same pitch as source (for algo)
					nextData = (BYTE *)malloc(height*src_pitch);
					next_created = true;
					env->BitBlt(nextData, src_pitch, pNext, next_pitch, rowsize, height);
					pNext = nextData;
				}

				if (norow)
				{
					if (mode==0)
						DeGrainMedian_YV12mode0norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==1)
						DeGrainMedian_YV12mode1norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==2)
						DeGrainMedian_YV12mode2norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==3)
						DeGrainMedian_YV12mode3norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==4)
						DeGrainMedian_YV12mode4norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==5)
						DeGrainMedian_YV12mode5norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
				}
				else
				{
					if (mode==0)
						DeGrainMedian_YV12mode0( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==1)
						DeGrainMedian_YV12mode1( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==2)
						DeGrainMedian_YV12mode2( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==3)
						DeGrainMedian_YV12mode3( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==4)
						DeGrainMedian_YV12mode4( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
					else if (mode==5)
						DeGrainMedian_YV12mode5( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitY);
				}
		

				// free created copy arrays
				if (prev_created == true) 
					free(prevData);
				if (next_created == true) 
					free(nextData);
			}
			else
				env->BitBlt(dstp, dst_pitch, srcp, src_pitch, rowsize, height);
			
			srcp = src->GetReadPtr(PLANAR_U);
			dstp = dst->GetWritePtr(PLANAR_U);
			src_pitch = src->GetPitch(PLANAR_U);
			dst_pitch = dst->GetPitch(PLANAR_U);
			rowsize = dst->GetRowSize(PLANAR_U);
			height = dst->GetHeight(PLANAR_U);
			if (limitUV>0 && useFrame >= 1 && useFrame < Max_Frames-1)
			{
				pPrev = PrevSrc->GetReadPtr(PLANAR_U);
				pNext = NextSrc->GetReadPtr(PLANAR_U);
				prev_pitch = PrevSrc->GetPitch(PLANAR_U);
				next_pitch = NextSrc->GetPitch(PLANAR_U);
				// check pitches and make its equal by copying (for algo) - added in v.0.4
				prev_created = false;
				if (prev_pitch != src_pitch)
				{ // copy frame data to array with same pitch as source 
					prevData = (BYTE *)malloc(height*src_pitch);
					prev_created = true;
					env->BitBlt(prevData, src_pitch, pPrev, prev_pitch, rowsize, height);
					pPrev = prevData;
				}
				next_created = false;
				if (next_pitch != src_pitch)
				{ // copy frame data to array with same pitch as source (for algo)
					nextData = (BYTE *)malloc(height*src_pitch);
					next_created = true;
					env->BitBlt(nextData, src_pitch, pNext, next_pitch, rowsize, height);
					pNext = nextData;
				}

				if (norow)
				{
					if (mode==0)
						DeGrainMedian_YV12mode0norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==1)
						DeGrainMedian_YV12mode1norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==2)
						DeGrainMedian_YV12mode2norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==3)
						DeGrainMedian_YV12mode3norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==4)
						DeGrainMedian_YV12mode4norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==5)
						DeGrainMedian_YV12mode5norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
				}
				else
				{
					if (mode==0)
						DeGrainMedian_YV12mode0( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==1)
						DeGrainMedian_YV12mode1( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==2)
						DeGrainMedian_YV12mode2( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==3)
						DeGrainMedian_YV12mode3( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==4)
						DeGrainMedian_YV12mode4( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==5)
						DeGrainMedian_YV12mode5( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
				}
		
				// free created copy arrays
				if (prev_created == true) 
					free(prevData);
				if (next_created == true) 
					free(nextData);
			}
			else
				env->BitBlt(dstp, dst_pitch, srcp, src_pitch, rowsize, height);
			
			srcp = src->GetReadPtr(PLANAR_V);
			dstp = dst->GetWritePtr(PLANAR_V);
			src_pitch = src->GetPitch(PLANAR_V);
			dst_pitch = dst->GetPitch(PLANAR_V);
			rowsize = dst->GetRowSize(PLANAR_V);
			height = dst->GetHeight(PLANAR_V);
			if (limitUV>0 && useFrame >= 1 && useFrame < Max_Frames-1)
			{
				pPrev = PrevSrc->GetReadPtr(PLANAR_V);
				pNext = NextSrc->GetReadPtr(PLANAR_V);
				prev_pitch = PrevSrc->GetPitch(PLANAR_V);
				next_pitch = NextSrc->GetPitch(PLANAR_V);
				// check pitches and make its equal by copying (for algo) - added in v.0.4
				prev_created = false;
				if (prev_pitch != src_pitch)
				{ // copy frame data to array with same pitch as source 
					prevData = (BYTE *)malloc(height*src_pitch);
					prev_created = true;
					env->BitBlt(prevData, src_pitch, pPrev, prev_pitch, rowsize, height);
					pPrev = prevData;
				}
				next_created = false;
				if (next_pitch != src_pitch)
				{ // copy frame data to array with same pitch as source (for algo)
					nextData = (BYTE *)malloc(height*src_pitch);
					next_created = true;
					env->BitBlt(nextData, src_pitch, pNext, next_pitch, rowsize, height);
					pNext = nextData;
				}

				if (norow)
				{
					if (mode==0)
						DeGrainMedian_YV12mode0norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==1)
						DeGrainMedian_YV12mode1norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==2)
						DeGrainMedian_YV12mode2norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==3)
						DeGrainMedian_YV12mode3norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==4)
						DeGrainMedian_YV12mode4norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==5)
						DeGrainMedian_YV12mode5norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
				}
				else
				{
					if (mode==0)
						DeGrainMedian_YV12mode0( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==1)
						DeGrainMedian_YV12mode1( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==2)
						DeGrainMedian_YV12mode2( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==3)
						DeGrainMedian_YV12mode3( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==4)
						DeGrainMedian_YV12mode4( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
					else if (mode==5)
						DeGrainMedian_YV12mode5( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext, rowsize, height, limitUV);
				}
		
				// free created copy arrays
				if (prev_created == true) 
					free(prevData);
				if (next_created == true) 
					free(nextData);
			}
			else
				env->BitBlt(dstp, dst_pitch, srcp, src_pitch, rowsize, height);
			
		}
		else // YUY2
		{
			srcp = src->GetReadPtr();
			dstp = dst->GetWritePtr();
			src_pitch = src->GetPitch();
			dst_pitch = dst->GetPitch();
			rowsize = dst->GetRowSize();
			height = dst->GetHeight();
			if ( useFrame >= 1 && useFrame < Max_Frames-1)
			{
				pPrev = PrevSrc->GetReadPtr();
				pNext = NextSrc->GetReadPtr();
				prev_pitch = PrevSrc->GetPitch();
				next_pitch = NextSrc->GetPitch();
				// check pitches and make its equal by copying (for algo) - added in v.0.4
				prev_created = false;
				if (prev_pitch != src_pitch)
				{ // copy frame data to array with same pitch as source 
					prevData = (BYTE *)malloc(height*src_pitch);
					prev_created = true;
					env->BitBlt(prevData, src_pitch, pPrev, prev_pitch, rowsize, height);
					pPrev = prevData;
				}
				next_created = false;
				if (next_pitch != src_pitch)
				{ // copy frame data to array with same pitch as source (for algo)
					nextData = (BYTE *)malloc(height*src_pitch);
					next_created = true;
					env->BitBlt(nextData, src_pitch, pNext, next_pitch, rowsize, height);
					pNext = nextData;
				}

				if (norow)
				{
					if (mode==0)
						DeGrainMedian_YUY2mode0norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext,	rowsize, height, limitY, limitUV);	
					else if (mode==1)
						DeGrainMedian_YUY2mode1norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext,	rowsize, height, limitY, limitUV);	
					else if (mode==2)
						DeGrainMedian_YUY2mode2norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext,	rowsize, height, limitY, limitUV);	
					else if (mode==3)
						DeGrainMedian_YUY2mode3norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext,	rowsize, height, limitY, limitUV);	
					else if (mode==4)
						DeGrainMedian_YUY2mode4norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext,	rowsize, height, limitY, limitUV);	
					else if (mode==5)
						DeGrainMedian_YUY2mode5norow( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext,	rowsize, height, limitY, limitUV);	
				}
				else
				{
					if (mode==0)
						DeGrainMedian_YUY2mode0( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext,	rowsize, height, limitY, limitUV);	
					else if (mode==1)
						DeGrainMedian_YUY2mode1( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext,	rowsize, height, limitY, limitUV);	
					else if (mode==2)
						DeGrainMedian_YUY2mode2( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext,	rowsize, height, limitY, limitUV);	
					else if (mode==3)
						DeGrainMedian_YUY2mode3( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext,	rowsize, height, limitY, limitUV);	
					else if (mode==4)
						DeGrainMedian_YUY2mode4( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext,	rowsize, height, limitY, limitUV);	
					else if (mode==5)
						DeGrainMedian_YUY2mode5( dstp,dst_pitch, pPrev, srcp, src_pitch, pNext,	rowsize, height, limitY, limitUV);	
				}
				// free created copy arrays
				if (prev_created == true) 
						free(prevData);
				if (next_created == true) 
						free(nextData);
			}
			else
				env->BitBlt(dstp, dst_pitch, srcp, src_pitch, rowsize, height);
		}

	}

	__asm emms;

    return dst;
}

//------------------------------------------------------------------------------------------------------

AVSValue __cdecl Create_DeGrainMedian(AVSValue args, void* user_data,
IScriptEnvironment* env) 
{
    return new DeGrainMedian(args[0].AsClip(),
		args[1].AsInt(4), // limitY
		args[2].AsInt(6), // limitUV
		args[3].AsInt(1), // mode
		args[4].AsBool(false), // interlaced
		args[5].AsBool(false), // norow
		env);
}


extern "C" __declspec(dllexport) const char* __stdcall
AvisynthPluginInit2(IScriptEnvironment* env) {
    env->AddFunction("DeGrainMedian", "c[limitY]i[limitUV]i[mode]i[interlaced]b[norow]b", Create_DeGrainMedian, 0);
    return "`DeGrainMedian' Spatio-Temporal Limited Median filter";
}

