/* 
DeGrainMedian plugin for Avisynth 2.5 - Spatio-Temporal Limited Median filter
Version 0.8 
Copyright (C)2004 Alexander G. Balakhnin aka Fizick.
bag@hotmail.ru         http://bag.hotmail.ru
under the GNU General Public Licence version 2.

 The plugin is based on STMedianFilter plugin by Tom Barry
 trbarry@trbarry.com
 Some ASM macros borrowed from RemoveGrain code by Rainer Wittmann <gorw@gmx.de> aka Kassandro :


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Change Log
2005/10/0 v0.8  renamed mode diagweight8 to diagweight8more and added true mode diagweight8
2004/09/10  v.0.4   Fixed possible bug with different frame pitches
2004/08/28  V 0.3	Added interlaced mode, code reorganization
2004/08/21  V 0.2	more safe modes1-4 is added 
2004/08/17  V 0.1.1	Bug fixed for spatial part of YV12 mode 
2004/08/13  V 0.1	Initial release 

*/


#ifndef __DEGRAINMEDIAN_H__
#define __DEGRAINMEDIAN_H__

#include "windows.h"
#include "avisynth.h"

const __int64 YMask   = 0x00ff00ff00ff00ff;
const __int64 UVMask  = 0xff00ff00ff00ff00;
const __int64 AllMask = 0xffffffffffffffff;


//------------------------------------------------------------------------------------------------------
// DeGrainMedian mode0 is from STMedianFilter (and similar to RemoveGrain mode9)

#define ChkBetterNeighbors(PADDR1, PADDR2) /* code by Tom Barry */\
__asm { /* input and output: mm5 - weigtht, mm4 - null, mm6 - min, mm7 - max */\
		__asm movq	mm0, qword ptr[(PADDR1)] /* our 4 (8) pixels */ \
		__asm movq mm1, qword ptr[(PADDR2)] /* our pixel2 value */ \
		__asm movq mm2, mm0			/* another copy of our pixel1 value */ \
		__asm pminub mm0, mm1       /* minimum value of the 2 */ \
		__asm pmaxub mm1, mm2       /* maximum value of the 2 */ \
		__asm movq mm2, mm1			/* another copy of our max value */ \
		__asm psubusb mm2, mm0 	    /* the diff */ \
		__asm movq mm3, mm2			/* another copy of our weights (diff) */ \
		__asm psubusb mm3, mm5		/* nonzero where old weights lower, else 0 */ \
		__asm pcmpeqb mm3, mm4		/* now ff where new better, else 00	*/ \
		__asm pand	mm0, mm3		/* keep only better min pixels */ \
		__asm pand	mm1, mm3		/* keep only better max pixels */ \
		__asm pand  mm2, mm3		/* and weights */ \
		__asm pcmpeqb mm3, mm4		/* here ff where old better, else 00 */ \
		__asm pand	mm5, mm3		/* keep only better old weights */ \
		__asm pand  mm6, mm3        /* keep only better old min  */ \
		__asm pand  mm7, mm3        /* keep only better old max  */ \
		__asm por	mm5, mm2		/* merge new & old weights */ \
		__asm por	mm6, mm0        /* merge new & old min */\
		__asm por	mm7, mm1		/* merge new & old max */ \
}

#define LimitPixelCorrection(LIMIT64) /* code by Fizick :-) */\
__asm { /* input: mm0 - original, mm4 - null,     output: mm0 - original, mm1 - limited result */\
		__asm movq	mm5, LIMIT64	/* correction limit */ \
		__asm movq	mm2, mm5		/* copy limit */ \
		__asm paddusb	mm2, mm0	/* max possible (mm0 is original) */ \
		__asm movq		mm3, mm1	/* (mm1 is changed) */ \
		__asm psubusb	mm3, mm0	/* changed - orig,   saturated to 0 */ \
		__asm psubusb	mm3, mm5	/* did it change too much, Y where nonzero */ \
		__asm pcmpeqb	mm3, mm4	/* now ff where new value should be used, else 00 (mm4=0)*/ \
		__asm pand	mm1, mm3		/* use new value for these pixels */\
		__asm pandn	mm3, mm2		/* use max value for these pixels */\
		__asm por	mm1, mm3		/* combine them, get result with limited  positive correction */\
		__asm movq	mm3, mm5		/* copy limit */\
		__asm movq	mm2, mm0		/* copy orig */\
		__asm psubusb	mm2, mm5	/* min possible */\
		__asm movq		mm3, mm0	/* copy orig */\
		__asm psubusb	mm3, mm1	/* orig - changed, saturated to 0 */\
		__asm psubusb	mm3, mm5	/* did it change too much, Y where nonzero */\
		__asm pcmpeqb	mm3, mm4	/* now ff where new value should be used, else 00 */\
		__asm pand	mm1, mm3		/* use new value for these pixels */\
		__asm pandn	mm3, mm2		/* use min value for these pixels */\
		__asm por	mm1, mm3		/* combine them, get result with limited  negative correction */\
}

// Macros borrowed from RemoveGrain code by Rainer Wittmann <gorw@gmx.de> aka Kassandro,
// released under GNU GPL.  In addition we deny anyone the right to derive a Photoshop plugin from the source code of RemoveGrain.

// DeGrainMedian mode0 is from STMedianFilter and similar to RemoveGrain mode9
// DeGrainMedian mode1 is from RemoveGrain mode8
// DeGrainMedian mode2 is from RemoveGrain mode7
// DeGrainMedian mode3 is from RemoveGrain mode6
// DeGrainMedian mode4 is from RemoveGrain mode5

// the values bound1 and bound2 are loaded into SSE registers
// then oldp is clipped with min(bound1, bound2) and max(bound1, bound2)
// finally weight = |oldp - newp|
// oldp is left unchanged
#define		diagweight5(oldp, newp, weight, bound1, bound2, reg1, reg2)	\
__asm	movq		newp,				bound1			\
__asm	movq		reg1,				bound2			\
__asm	movq		weight,				newp			\
__asm	movq		reg2,				oldp			\
__asm	pmaxub		newp,				reg1			\
__asm	pminub		weight,				reg1			\
__asm	psubusb		reg2,				newp			\
__asm	pminub		newp,				oldp			\
__asm	pmaxub		newp,				weight			\
__asm	psubusb		weight,				oldp			\
__asm	pmaxub		weight,				reg2

// the values bound1 and bound2 are loaded into SSE registers
// then oldp is clipped with min(bound1, bound2) and max(bound1, bound2)
// finally weight = 2*|oldp - newp| + |bound1 - bound2|
// oldp is left unchanged
#define		diagweight6(oldp, newp, weight, bound1, bound2, reg1, reg2)	\
__asm	movq		newp,				bound1			\
__asm	movq		reg1,				bound2			\
__asm	movq		weight,				newp			\
__asm	movq		reg2,				oldp			\
__asm	pmaxub		newp,				reg1			\
__asm	pminub		weight,				reg1			\
__asm	psubusb		reg2,				newp			\
__asm	movq		reg1,				newp			\
__asm	pminub		newp,				oldp			\
__asm	psubusb		reg1,				weight			\
__asm	pmaxub		newp,				weight			\
__asm	psubusb		weight,				oldp			\
__asm	pmaxub		weight,				reg2			\
__asm	paddusb		weight,				weight			\
__asm	paddusb		weight,				reg1

// the values bound1 and bound2 are loaded into SSE registers
// then oldp is clipped with min(bound1, bound2) and max(bound1, bound2)
// finally weight = |oldp - newp| + |bound1 - bound2|
// oldp is left unchanged
#define		diagweight7(oldp, newp, weight, bound1, bound2, reg1, reg2)	\
__asm	movq		newp,				bound1			\
__asm	movq		reg1,				bound2			\
__asm	movq		weight,				newp			\
__asm	movq		reg2,				oldp			\
__asm	pmaxub		newp,				reg1			\
__asm	pminub		weight,				reg1			\
__asm	psubusb		reg2,				newp			\
__asm	movq		reg1,				newp			\
__asm	pminub		newp,				oldp			\
__asm	psubusb		reg1,				weight			\
__asm	pmaxub		newp,				weight			\
__asm	psubusb		weight,				oldp			\
__asm	pmaxub		weight,				reg2			\
__asm	paddusb		weight,				reg1


// the values bound1 and bound2 are loaded into SSE registers
// then oldp is clipped with min(bound1, bound2) and max(bound1, bound2)
// finally weight = |oldp - newp| + 2*|bound1 - bound2|
// oldp is left unchanged
#define		diagweight8(oldp, newp, weight, bound1, bound2, reg1, reg2)	\
__asm	movq		newp,				bound1			\
__asm	movq		reg1,				bound2			\
__asm	movq		weight,				newp /*bound1 */			\
__asm	movq		reg2,				oldp			\
__asm	pmaxub		newp,				reg1	/* max(bound1, bound2 */		\
__asm	pminub		weight,				reg1	/* min(bound1, bound2) */		\
__asm	psubusb		reg2,				newp	/* oldp - max(bound1,bound2) */		\
__asm	movq		reg1,				newp	/* max(bound1, bound2) */		\
__asm	pminub		newp,				oldp	/* min(max(bound1, bound2), oldp) */		\
__asm	psubusb		reg1,				weight	/* max(bound1, bound2) - min(bound1, bound2) = |bound1-bound2|*/ \
__asm	pmaxub		newp,				weight	/* max( min(max(bound1, bound2), oldp), min(bound1, bound2) )*/		\
__asm	psubusb		weight,				oldp	/* min(bound1, bound2) - oldp */		\
__asm	paddusb		reg1,				reg1	/* |bound1-bound2|*2	*/	\
__asm	pmaxub		weight,				reg2	/* max(min(bound1, bound2) - oldp), oldp - max(bound1,bound2) )	*/	\
__asm	paddusb		weight,				reg1    /* weight + |bound1-bound2|*2  */

// the values bound1 and bound2 are loaded into SSE registers
// then oldp is clipped with min(bound1, bound2) and max(bound1, bound2)
// finally weight = |oldp - newp| + 4*|bound1 - bound2|
// oldp is left unchanged
#define		diagweight8more(oldp, newp, weight, bound1, bound2, reg1, reg2)	\
__asm	movq		newp,				bound1			\
__asm	movq		reg1,				bound2			\
__asm	movq		weight,				newp /*bound1 */			\
__asm	movq		reg2,				oldp			\
__asm	pmaxub		newp,				reg1	/* max(bound1, bound2 */		\
__asm	pminub		weight,				reg1	/* min(bound1, bound2) */		\
__asm	psubusb		reg2,				newp	/* oldp - max(bound1,bound2) */		\
__asm	movq		reg1,				newp	/* max(bound1, bound2) */		\
__asm	pminub		newp,				oldp	/* min(max(bound1, bound2), oldp) */		\
__asm	psubusb		reg1,				weight	/* max(bound1, bound2) - min(bound1, bound2) = |bound1-bound2|*/ \
__asm	pmaxub		newp,				weight	/* max( min(max(bound1, bound2), oldp), min(bound1, bound2) )*/		\
__asm	psubusb		weight,				oldp	/* min(bound1, bound2) - oldp */		\
__asm	paddusb		reg1,				reg1	/* |bound1-bound2|*2	*/	\
__asm	paddusb		reg1,				reg1	/* |bound1-bound2|*4	*/	\
__asm	pmaxub		weight,				reg2	/* max(min(bound1, bound2) - oldp), oldp - max(bound1,bound2) )	*/	\
__asm	paddusb		weight,				reg1    /* weight + |bound1-bound2|*4  */

// if( weight2[i] <= weight1[i] ) { value1[i] = value2[i]; weight1[i] = weight2[i]; }
// value2 remains unchanged
// weight2 must be a SSE register, value1, value2, weight1 may very well be a memory variables
// but value1 and weight1 should be registers because they are used twice
#define	mergeweighted(value1, weight1, value2, weight2) \
__asm	pminub	weight1, weight2 /* w1 = min(weight1, weight2) */ \
__asm	pcmpeqb	weight2, weight1 /* w2 = cmp w2, min(w1,w2) : =FF if w2 was min , =0 if not */\
__asm	psubusb	value1, weight2 /* value1 = 0 if above W2=FF, i.e. if w2 was min, = old v1 if not */\
__asm	pand	weight2, value2 /* weigth2 = value2 if w2 was min , =0 if not */\
__asm	por		value1, weight2 /* */

#endif