/*
DeGrainMedian plugin for Avisynth 2.5 - Spatio-Temporal Limited Median filter
Version 0.8 
Copyright (C)2004 Alexander G. Balakhnin aka Fizick.
bag@hotmail.ru         http://bag.hotmail.ru
under the GNU General Public Licence version 2.

 The plugin is based on STMedianFilter plugin by Tom Barry
 trbarry@trbarry.com
 Some ASM macros borrowed from RemoveGrain code by Rainer Wittmann <gorw@gmx.de> aka Kassandro :

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Change Log
2005/10/08 - renamed mode 2 to 3, 3 to 4, 4 to 5, added mode 2
             added norow functions
2004/09/10  v.0.4   Fixed possible bug with different frame pitches
2004/08/28  V 0.3	Added interlaced mode, code reorganization
2004/08/21  V 0.2.0.0	more safe modes1-4 is added 
2004/08/17  V 0.1.1.0	Bug fixed for spatial part of YV12 mode 
2004/08/13  V 0.1.0.0	Initial release 

*/

#include "windows.h"
#include "avisynth.h"
#include "degrainmedian.h"

//--------------------------------------------------------------------
//  YV12 code
//--------------------------------------------------------------------
// Same function as above except for offset to adjacent pixels and chroma mask

void DeGrainMedian_YV12mode0(BYTE* dstp, int dst_pitch, const BYTE* _pPrev,
			const BYTE* srcp, int src_pitch, const BYTE* _pNext,
			int rowsize, int FldHeight, int Limit)				
{

const BYTE* pPrev = _pPrev;
const BYTE* pNext = _pNext;
const BYTE* pSrc = srcp;

BYTE* pDest = dstp;
int y;
int ct = ((rowsize-9) >> 3); // changed from -10 to -9 (chroma may be odd) in v0.8.2 
int lastback = 8 + (ct<<3) + 8 - rowsize; // v0.8.2
__int64	i = Limit;	
__int64	LimitB = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 

	// first and last line are not filtered
	memcpy(pDest, pSrc, rowsize);		// copy first line
	memcpy(pDest+dst_pitch*(FldHeight-1),	pSrc+src_pitch*(FldHeight-1), rowsize);  // and last

	pDest += dst_pitch; // go to second line

	for (y=1; y <= FldHeight-2; y++)	
	{

		__asm
		{
		// Loop general reg usage
		//
		// eax - inner loop counter
		// ebx - src lines above curr 
		// ecx - src pitch
		// edx - dest
		// edi - src pixels, next frame, 1 line up
		// esi - src pixels, prev frame, 1 line up

		mov		ebx, pSrc   
		mov		esi, pPrev
		mov		edi, pNext
		mov		ecx, src_pitch; //src_pitch 
		mov		edx, pDest 

		pxor	mm4, mm4						// useful constant 000..	


		//	do some filtering of first 8 pixels  
		pcmpeqb	mm5, mm5		// init diff 255's
		pxor	mm6, mm6		// init min to 0
		pcmpeqb	mm7, mm7		// init max 255's

// Search between prev and next frame - Temporal
// The same row only
        ChkBetterNeighbors(edi+2*ecx, esi)		// down, up
        ChkBetterNeighbors(edi, esi+2*ecx)		// up, down
		ChkBetterNeighbors(edi+ecx, esi+ecx)	// center, in old and new
// Spatial
        ChkBetterNeighbors(ebx, ebx+2*ecx)	// up, down

// We now know the min and max of the closest together pair of neighbors
		movq	mm0, qword ptr[ebx+ecx]		// our curr pixels vals
		movq	mm1, mm0
		pminub	mm1, mm7					// make it no larger than max neighbor
		pmaxub	mm1, mm6					// and now smaller than min neighbor

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
		// make some adjustments to start output on 2nd qword of 2nd line
		add		edi, 8
		add		ebx, 8
		add		esi, 8
		add		edx, 8
		mov	    eax, ct			// loop ct

LoopQ:	
		pcmpeqb	mm5, mm5		// init diff 255's
		pxor	mm6, mm6		// init min to 0
		pcmpeqb	mm7, mm7		// init max 255's

// Search between prev and next frame - Temporal
        ChkBetterNeighbors(edi-1, esi+2*ecx+1)  // up left, down right
        ChkBetterNeighbors(edi+1, esi+2*ecx-1)  // up right, down left
        ChkBetterNeighbors(edi+2*ecx-1, esi+1)  // down left, up right
        ChkBetterNeighbors(edi+2*ecx+1, esi-1)   // down right, up left   
        ChkBetterNeighbors(edi+2*ecx, esi)		// down, up
        ChkBetterNeighbors(edi, esi+2*ecx)		// up, down
        ChkBetterNeighbors(edi+ecx-1, esi+ecx+1) // left, right
        ChkBetterNeighbors(edi+ecx+1, esi+ecx-1) // right, left
		ChkBetterNeighbors(edi+ecx, esi+ecx)	// center, in old and new
// Spatial
        ChkBetterNeighbors(ebx-1, ebx+2*ecx+1)  // up left, down right
        ChkBetterNeighbors(ebx+1, ebx+2*ecx-1)  // up right, down left
        ChkBetterNeighbors(ebx, ebx+2*ecx)	// up, down
        ChkBetterNeighbors(ebx+ecx-1, ebx+ecx+1) // left, right 

// We now know the min and max of the closest together pair of neighbors
		movq	mm0, qword ptr[ebx+ecx]		// our curr pixels vals
		movq	mm1, mm0
		pminub	mm1, mm7					// make it no larger than max neighbor
		pmaxub	mm1, mm6					// and now smaller than min neighbor

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
		add		esi, 8
		add		edi, 8
		add		ebx, 8
		add		edx, 8

		dec		eax
		jnz		LoopQ

		//	do some filtering of last 8 pixels
		// rewind back to set last column equal to row size - v0.8.2
		mov eax, lastback 
		sub	esi, eax
		sub	edi, eax
		sub ebx, eax
		sub edx, eax

		pcmpeqb	mm5, mm5		// init diff 255's
		pxor	mm6, mm6		// init min to 0
		pcmpeqb	mm7, mm7		// init max 255's

// Search between prev and next frame - Temporal
		// The same row only
        ChkBetterNeighbors(edi+2*ecx, esi)		// down, up
        ChkBetterNeighbors(edi, esi+2*ecx)		// up, down
		ChkBetterNeighbors(edi+ecx, esi+ecx)	// center, in old and new
// Spatial
        ChkBetterNeighbors(ebx, ebx+2*ecx)	// up, down

// We now know the min and max of the closest together pair of neighbors
		movq	mm0, qword ptr[ebx+ecx]		// our curr pixels vals
		movq	mm1, mm0
		pminub	mm1, mm7					// make it no larger than max neighbor
		pmaxub	mm1, mm6					// and now smaller than min neighbor

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
	}

	// adjust for next line
	pSrc  += src_pitch;
	pPrev  += src_pitch;
	pNext  += src_pitch;
	pDest += dst_pitch;
	}
}

//---------------------------------------------------------------------------------------------
// Same function as above except for mode
void DeGrainMedian_YV12mode1(BYTE* dstp, int dst_pitch, const BYTE* _pPrev,
			const BYTE* srcp, int src_pitch, const BYTE* _pNext,
			int rowsize, int FldHeight, int Limit)				
{

const BYTE* pPrev = _pPrev;
const BYTE* pNext = _pNext;
const BYTE* pSrc = srcp;

BYTE* pDest = dstp;
int y;
int ct = ((rowsize-9) >> 3);
int lastback = 8 + (ct<<3) + 8 - rowsize; // v0.8.2
__int64	i = Limit;	
__int64	LimitB = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 

	// first and last line are not filtered
	memcpy(pDest, pSrc, rowsize);		// copy first line
	memcpy(pDest+dst_pitch*(FldHeight-1),	pSrc+src_pitch*(FldHeight-1), rowsize);  // and last

	pDest += dst_pitch; // go to second line

	for (y=1; y <= FldHeight-2; y++)	
	{

		__asm
		{
		// Loop general reg usage
		//
		// eax - inner loop counter
		// ebx - src lines above curr 
		// ecx - src pitch
		// edx - dest
		// edi - src pixels, next frame, 1 line up
		// esi - src pixels, prev frame, 1 line up

		mov		ebx, pSrc   
		mov		esi, pPrev
		mov		edi, pNext
		mov		ecx, src_pitch; //src_pitch 
		mov		edx, pDest 

//		pxor	mm4, mm4						// useful constant 000..	


		//	do some filtering of first 8 pixels  

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// Search Spatial
		diagweight8more(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight8more(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
		// make some adjustments to start output on 2nd qword of 2nd line
		add		edi, 8
		add		ebx, 8
		add		esi, 8
		add		edx, 8
		mov	    eax, ct			// loop ct

LoopQ:	

// Search between prev and next frame - Temporal

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// diagweight8(oldp, newp, weight, bound1, bound2, reg1, reg2)	
// oldp is clipped with min(bound1, bound2) and max(bound1, bound2)
// finally weight = |oldp - newp| + 2*|bound1 - bound2|
// Search Spatial
		diagweight8more(mm0, mm1, mm2, [ebx - 1], [ebx + 2*ecx + 1], mm6, mm7) // up left, down right
		diagweight8more(mm0, mm3, mm4, [ebx + 2*ecx -1], [ebx + 1], mm6, mm7)// up right, down left
//	mergeweighted(value1, weight1, value2, weight2) 
// if( weight2[i] <= weight1[i] ) { value1[i] = value2[i]; weight1[i] = weight2[i]; }
// value2 remains unchanged
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [ebx + ecx + 1], [ebx + ecx - 1], mm6, mm7)// left, right 
		mergeweighted(mm1, mm2, mm3, mm4)
// Search between prev and next frame - Temporal
		diagweight8more(mm0, mm3, mm4, [edi-1], [esi+2*ecx+1], mm6, mm7)// up left, down right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+1], [esi+2*ecx-1], mm6, mm7)// up right, down left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+2*ecx-1], [esi+1], mm6, mm7)// down left, up right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+2*ecx+1], [esi-1], mm6, mm7)// down right, up left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+ecx-1], [esi+ecx+1], mm6, mm7)// left, right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+ecx+1], [esi+ecx-1], mm6, mm7)// right, left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..
		
		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
		add		esi, 8
		add		edi, 8
		add		ebx, 8
		add		edx, 8

		dec		eax
		jnz		LoopQ

		//	do some filtering of last 8 pixels  
		// rewind back to set last column equal to row size - v0.8.2
		mov eax, lastback 
		sub	esi, eax
		sub	edi, eax
		sub ebx, eax
		sub edx, eax
		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

		// Search Spatial
		diagweight8more(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight8more(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
	}

	// adjust for next line
	pSrc  += src_pitch;
	pPrev  += src_pitch;
	pNext  += src_pitch;
	pDest += dst_pitch;
	}
}

//---------------------------------------------------------------------------------------------
// Same function as above except for mode
void DeGrainMedian_YV12mode2(BYTE* dstp, int dst_pitch, const BYTE* _pPrev,
			const BYTE* srcp, int src_pitch, const BYTE* _pNext,
			int rowsize, int FldHeight, int Limit)				
{

const BYTE* pPrev = _pPrev;
const BYTE* pNext = _pNext;
const BYTE* pSrc = srcp;

BYTE* pDest = dstp;
int y;
int ct = ((rowsize-9) >> 3);
int lastback = 8 + (ct<<3) + 8 - rowsize; // v0.8.2
__int64	i = Limit;	
__int64	LimitB = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 

	// first and last line are not filtered
	memcpy(pDest, pSrc, rowsize);		// copy first line
	memcpy(pDest+dst_pitch*(FldHeight-1),	pSrc+src_pitch*(FldHeight-1), rowsize);  // and last

	pDest += dst_pitch; // go to second line

	for (y=1; y <= FldHeight-2; y++)	
	{

		__asm
		{
		// Loop general reg usage
		//
		// eax - inner loop counter
		// ebx - src lines above curr 
		// ecx - src pitch
		// edx - dest
		// edi - src pixels, next frame, 1 line up
		// esi - src pixels, prev frame, 1 line up

		mov		ebx, pSrc   
		mov		esi, pPrev
		mov		edi, pNext
		mov		ecx, src_pitch; //src_pitch 
		mov		edx, pDest 

//		pxor	mm4, mm4						// useful constant 000..	


		//	do some filtering of first 8 pixels  

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// Search Spatial
		diagweight8(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight8(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
		// make some adjustments to start output on 2nd qword of 2nd line
		add		edi, 8
		add		ebx, 8
		add		esi, 8
		add		edx, 8
		mov	    eax, ct			// loop ct

LoopQ:	

// Search between prev and next frame - Temporal

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// diagweight8(oldp, newp, weight, bound1, bound2, reg1, reg2)	
// oldp is clipped with min(bound1, bound2) and max(bound1, bound2)
// finally weight = |oldp - newp| + 2*|bound1 - bound2|
// Search Spatial
		diagweight8(mm0, mm1, mm2, [ebx - 1], [ebx + 2*ecx + 1], mm6, mm7) // up left, down right
		diagweight8(mm0, mm3, mm4, [ebx + 2*ecx -1], [ebx + 1], mm6, mm7)// up right, down left
//	mergeweighted(value1, weight1, value2, weight2) 
// if( weight2[i] <= weight1[i] ) { value1[i] = value2[i]; weight1[i] = weight2[i]; }
// value2 remains unchanged
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [ebx + ecx + 1], [ebx + ecx - 1], mm6, mm7)// left, right 
		mergeweighted(mm1, mm2, mm3, mm4)
// Search between prev and next frame - Temporal
		diagweight8(mm0, mm3, mm4, [edi-1], [esi+2*ecx+1], mm6, mm7)// up left, down right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+1], [esi+2*ecx-1], mm6, mm7)// up right, down left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+2*ecx-1], [esi+1], mm6, mm7)// down left, up right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+2*ecx+1], [esi-1], mm6, mm7)// down right, up left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+ecx-1], [esi+ecx+1], mm6, mm7)// left, right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+ecx+1], [esi+ecx-1], mm6, mm7)// right, left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..
		
		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
		add		esi, 8
		add		edi, 8
		add		ebx, 8
		add		edx, 8

		dec		eax
		jnz		LoopQ

		//	do some filtering of last 8 pixels  
		// rewind back to set last column equal to row size - v0.8.2
		mov eax, lastback 
		sub	esi, eax
		sub	edi, eax
		sub ebx, eax
		sub edx, eax
		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

		// Search Spatial
		diagweight8(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight8(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
	}

	// adjust for next line
	pSrc  += src_pitch;
	pPrev  += src_pitch;
	pNext  += src_pitch;
	pDest += dst_pitch;
	}
}

//---------------------------------------------------------------------------------------------
// Same function as above except for mode
void DeGrainMedian_YV12mode3(BYTE* dstp, int dst_pitch, const BYTE* _pPrev,
			const BYTE* srcp, int src_pitch, const BYTE* _pNext,
			int rowsize, int FldHeight, int Limit)				
{

const BYTE* pPrev = _pPrev;
const BYTE* pNext = _pNext;
const BYTE* pSrc = srcp;

BYTE* pDest = dstp;
int y;
int ct = ((rowsize-9) >> 3);
int lastback = 8 + (ct<<3) + 8 - rowsize; // v0.8.2
__int64	i = Limit;	
__int64	LimitB = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 

	// first and last line are not filtered
	memcpy(pDest, pSrc, rowsize);		// copy first line
	memcpy(pDest+dst_pitch*(FldHeight-1),	pSrc+src_pitch*(FldHeight-1), rowsize);  // and last

	pDest += dst_pitch; // go to second line

	for (y=1; y <= FldHeight-2; y++)	
	{

		__asm
		{
		// Loop general reg usage
		//
		// eax - inner loop counter
		// ebx - src lines above curr 
		// ecx - src pitch
		// edx - dest
		// edi - src pixels, next frame, 1 line up
		// esi - src pixels, prev frame, 1 line up

		mov		ebx, pSrc   
		mov		esi, pPrev
		mov		edi, pNext
		mov		ecx, src_pitch; //src_pitch 
		mov		edx, pDest 

//		pxor	mm4, mm4						// useful constant 000..	


		//	do some filtering of first 8 pixels  

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// Search Spatial
		diagweight7(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight7(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
		// make some adjustments to start output on 2nd qword of 2nd line
		add		edi, 8
		add		ebx, 8
		add		esi, 8
		add		edx, 8
		mov	    eax, ct			// loop ct

LoopQ:	

// Search between prev and next frame - Temporal

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// diagweight7(oldp, newp, weight, bound1, bound2, reg1, reg2)	
// oldp is clipped with min(bound1, bound2) and max(bound1, bound2)
// finally weight = |oldp - newp| + 2*|bound1 - bound2|
// Search Spatial
		diagweight7(mm0, mm1, mm2, [ebx - 1], [ebx + 2*ecx + 1], mm6, mm7) // up left, down right
		diagweight7(mm0, mm3, mm4, [ebx + 2*ecx -1], [ebx + 1], mm6, mm7)// up right, down left
//	mergeweighted(value1, weight1, value2, weight2) 
// if( weight2[i] <= weight1[i] ) { value1[i] = value2[i]; weight1[i] = weight2[i]; }
// value2 remains unchanged
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [ebx + ecx + 1], [ebx + ecx - 1], mm6, mm7)// left, right 
		mergeweighted(mm1, mm2, mm3, mm4)
// Search between prev and next frame - Temporal
		diagweight7(mm0, mm3, mm4, [edi-1], [esi+2*ecx+1], mm6, mm7)// up left, down right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+1], [esi+2*ecx-1], mm6, mm7)// up right, down left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+2*ecx-1], [esi+1], mm6, mm7)// down left, up right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+2*ecx+1], [esi-1], mm6, mm7)// down right, up left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+ecx-1], [esi+ecx+1], mm6, mm7)// left, right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+ecx+1], [esi+ecx-1], mm6, mm7)// right, left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..
		
		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
		add		esi, 8
		add		edi, 8
		add		ebx, 8
		add		edx, 8

		dec		eax
		jnz		LoopQ

		//	do some filtering of last 8 pixels  
		// rewind back to set last column equal to row size - v0.8.2
		mov eax, lastback 
		sub	esi, eax
		sub	edi, eax
		sub ebx, eax
		sub edx, eax
		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

		// Search Spatial
		diagweight7(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight7(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
	}

	// adjust for next line
	pSrc  += src_pitch;
	pPrev  += src_pitch;
	pNext  += src_pitch;
	pDest += dst_pitch;
	}
}

//---------------------------------------------------------------------------------------------
// Same function as above except for mode
void DeGrainMedian_YV12mode4(BYTE* dstp, int dst_pitch, const BYTE* _pPrev,
			const BYTE* srcp, int src_pitch, const BYTE* _pNext,
			int rowsize, int FldHeight, int Limit)				
{

const BYTE* pPrev = _pPrev;
const BYTE* pNext = _pNext;
const BYTE* pSrc = srcp;

BYTE* pDest = dstp;
int y;
int ct = ((rowsize-9) >> 3);
int lastback = 8 + (ct<<3) + 8 - rowsize; // v0.8.2
__int64	i = Limit;	
__int64	LimitB = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 

	// first and last line are not filtered
	memcpy(pDest, pSrc, rowsize);		// copy first line
	memcpy(pDest+dst_pitch*(FldHeight-1),	pSrc+src_pitch*(FldHeight-1), rowsize);  // and last

	pDest += dst_pitch; // go to second line

	for (y=1; y <= FldHeight-2; y++)	
	{

		__asm
		{
		// Loop general reg usage
		//
		// eax - inner loop counter
		// ebx - src lines above curr 
		// ecx - src pitch
		// edx - dest
		// edi - src pixels, next frame, 1 line up
		// esi - src pixels, prev frame, 1 line up

		mov		ebx, pSrc   
		mov		esi, pPrev
		mov		edi, pNext
		mov		ecx, src_pitch; //src_pitch 
		mov		edx, pDest 

//		pxor	mm4, mm4						// useful constant 000..	


		//	do some filtering of first 8 pixels  

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// Search Spatial
		diagweight6(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight6(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
		// make some adjustments to start output on 2nd qword of 2nd line
		add		edi, 8
		add		ebx, 8
		add		esi, 8
		add		edx, 8
		mov	    eax, ct			// loop ct

LoopQ:	

// Search between prev and next frame - Temporal

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// diagweight6(oldp, newp, weight, bound1, bound2, reg1, reg2)	
// oldp is clipped with min(bound1, bound2) and max(bound1, bound2)
// finally weight = |oldp - newp| + 2*|bound1 - bound2|
// Search Spatial
		diagweight6(mm0, mm1, mm2, [ebx - 1], [ebx + 2*ecx + 1], mm6, mm7) // up left, down right
		diagweight6(mm0, mm3, mm4, [ebx + 2*ecx -1], [ebx + 1], mm6, mm7)// up right, down left
//	mergeweighted(value1, weight1, value2, weight2) 
// if( weight2[i] <= weight1[i] ) { value1[i] = value2[i]; weight1[i] = weight2[i]; }
// value2 remains unchanged
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [ebx + ecx + 1], [ebx + ecx - 1], mm6, mm7)// left, right 
		mergeweighted(mm1, mm2, mm3, mm4)
// Search between prev and next frame - Temporal
		diagweight6(mm0, mm3, mm4, [edi-1], [esi+2*ecx+1], mm6, mm7)// up left, down right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+1], [esi+2*ecx-1], mm6, mm7)// up right, down left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+2*ecx-1], [esi+1], mm6, mm7)// down left, up right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+2*ecx+1], [esi-1], mm6, mm7)// down right, up left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+ecx-1], [esi+ecx+1], mm6, mm7)// left, right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+ecx+1], [esi+ecx-1], mm6, mm7)// right, left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..
		
		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
		add		esi, 8
		add		edi, 8
		add		ebx, 8
		add		edx, 8

		dec		eax
		jnz		LoopQ

		//	do some filtering of last 8 pixels  
		// rewind back to set last column equal to row size - v0.8.2
		mov eax, lastback 
		sub	esi, eax
		sub	edi, eax
		sub ebx, eax
		sub edx, eax
		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

		// Search Spatial
		diagweight6(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight6(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
	}

	// adjust for next line
	pSrc  += src_pitch;
	pPrev  += src_pitch;
	pNext  += src_pitch;
	pDest += dst_pitch;
	}
}

//---------------------------------------------------------------------------------------------
// Same function as above except for mode
void DeGrainMedian_YV12mode5(BYTE* dstp, int dst_pitch, const BYTE* _pPrev,
			const BYTE* srcp, int src_pitch, const BYTE* _pNext,
			int rowsize, int FldHeight, int Limit)				
{

const BYTE* pPrev = _pPrev;
const BYTE* pNext = _pNext;
const BYTE* pSrc = srcp;

BYTE* pDest = dstp;
int y;
int ct = ((rowsize-9) >> 3);
int lastback = 8 + (ct<<3) + 8 - rowsize; // v0.8.2
__int64	i = Limit;	
__int64	LimitB = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 

	// first and last line are not filtered
	memcpy(pDest, pSrc, rowsize);		// copy first line
	memcpy(pDest+dst_pitch*(FldHeight-1),	pSrc+src_pitch*(FldHeight-1), rowsize);  // and last

	pDest += dst_pitch; // go to second line

	for (y=1; y <= FldHeight-2; y++)	
	{

		__asm
		{
		// Loop general reg usage
		//
		// eax - inner loop counter
		// ebx - src lines above curr 
		// ecx - src pitch
		// edx - dest
		// edi - src pixels, next frame, 1 line up
		// esi - src pixels, prev frame, 1 line up

		mov		ebx, pSrc   
		mov		esi, pPrev
		mov		edi, pNext
		mov		ecx, src_pitch; //src_pitch 
		mov		edx, pDest 

//		pxor	mm4, mm4						// useful constant 000..	


		//	do some filtering of first 8 pixels  

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// Search Spatial
		diagweight5(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight5(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
		// make some adjustments to start output on 2nd qword of 2nd line
		add		edi, 8
		add		ebx, 8
		add		esi, 8
		add		edx, 8
		mov	    eax, ct			// loop ct

LoopQ:	

// Search between prev and next frame - Temporal

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// diagweight5(oldp, newp, weight, bound1, bound2, reg1, reg2)	
// oldp is clipped with min(bound1, bound2) and max(bound1, bound2)
// finally weight = |oldp - newp| + 2*|bound1 - bound2|
// Search Spatial
		diagweight5(mm0, mm1, mm2, [ebx - 1], [ebx + 2*ecx + 1], mm6, mm7) // up left, down right
		diagweight5(mm0, mm3, mm4, [ebx + 2*ecx -1], [ebx + 1], mm6, mm7)// up right, down left
//	mergeweighted(value1, weight1, value2, weight2) 
// if( weight2[i] <= weight1[i] ) { value1[i] = value2[i]; weight1[i] = weight2[i]; }
// value2 remains unchanged
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [ebx + ecx + 1], [ebx + ecx - 1], mm6, mm7)// left, right 
		mergeweighted(mm1, mm2, mm3, mm4)
// Search between prev and next frame - Temporal
		diagweight5(mm0, mm3, mm4, [edi-1], [esi+2*ecx+1], mm6, mm7)// up left, down right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+1], [esi+2*ecx-1], mm6, mm7)// up right, down left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+2*ecx-1], [esi+1], mm6, mm7)// down left, up right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+2*ecx+1], [esi-1], mm6, mm7)// down right, up left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+ecx-1], [esi+ecx+1], mm6, mm7)// left, right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+ecx+1], [esi+ecx-1], mm6, mm7)// right, left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..
		
		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
		add		esi, 8
		add		edi, 8
		add		ebx, 8
		add		edx, 8

		dec		eax
		jnz		LoopQ

		//	do some filtering of last 8 pixels  
		// rewind back to set last column equal to row size - v0.8.2
		mov eax, lastback 
		sub	esi, eax
		sub	edi, eax
		sub ebx, eax
		sub edx, eax
		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

		// Search Spatial
		diagweight5(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight5(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
	}

	// adjust for next line
	pSrc  += src_pitch;
	pPrev  += src_pitch;
	pNext  += src_pitch;
	pDest += dst_pitch;
	}
}

//------------------------------------------------------------------------------------------------------
// NO ROW FUCTIONS (do not use same row in spatial filter)
//------------------------------------------------------------------------------------------------------
void DeGrainMedian_YV12mode0norow(BYTE* dstp, int dst_pitch, const BYTE* _pPrev,
			const BYTE* srcp, int src_pitch, const BYTE* _pNext,
			int rowsize, int FldHeight, int Limit)				
{

const BYTE* pPrev = _pPrev;
const BYTE* pNext = _pNext;
const BYTE* pSrc = srcp;

BYTE* pDest = dstp;
int y;
int ct = ((rowsize-9) >> 3);
int lastback = 8 + (ct<<3) + 8 - rowsize; // v0.8.2
__int64	i = Limit;	
__int64	LimitB = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 

	// first and last line are not filtered
	memcpy(pDest, pSrc, rowsize);		// copy first line
	memcpy(pDest+dst_pitch*(FldHeight-1),	pSrc+src_pitch*(FldHeight-1), rowsize);  // and last

	pDest += dst_pitch; // go to second line

	for (y=1; y <= FldHeight-2; y++)	
	{

		__asm
		{
		// Loop general reg usage
		//
		// eax - inner loop counter
		// ebx - src lines above curr 
		// ecx - src pitch
		// edx - dest
		// edi - src pixels, next frame, 1 line up
		// esi - src pixels, prev frame, 1 line up

		mov		ebx, pSrc   
		mov		esi, pPrev
		mov		edi, pNext
		mov		ecx, src_pitch; //src_pitch 
		mov		edx, pDest 

		pxor	mm4, mm4						// useful constant 000..	


		//	do some filtering of first 8 pixels  
		pcmpeqb	mm5, mm5		// init diff 255's
		pxor	mm6, mm6		// init min to 0
		pcmpeqb	mm7, mm7		// init max 255's

// Search between prev and next frame - Temporal
// The same row only
        ChkBetterNeighbors(edi+2*ecx, esi)		// down, up
        ChkBetterNeighbors(edi, esi+2*ecx)		// up, down
		ChkBetterNeighbors(edi+ecx, esi+ecx)	// center, in old and new
// Spatial
        ChkBetterNeighbors(ebx, ebx+2*ecx)	// up, down

// We now know the min and max of the closest together pair of neighbors
		movq	mm0, qword ptr[ebx+ecx]		// our curr pixels vals
		movq	mm1, mm0
		pminub	mm1, mm7					// make it no larger than max neighbor
		pmaxub	mm1, mm6					// and now smaller than min neighbor

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
		// make some adjustments to start output on 2nd qword of 2nd line
		add		edi, 8
		add		ebx, 8
		add		esi, 8
		add		edx, 8
		mov	    eax, ct			// loop ct

LoopQ:	
		pcmpeqb	mm5, mm5		// init diff 255's
		pxor	mm6, mm6		// init min to 0
		pcmpeqb	mm7, mm7		// init max 255's

// Search between prev and next frame - Temporal
        ChkBetterNeighbors(edi-1, esi+2*ecx+1)  // up left, down right
        ChkBetterNeighbors(edi+1, esi+2*ecx-1)  // up right, down left
        ChkBetterNeighbors(edi+2*ecx-1, esi+1)  // down left, up right
        ChkBetterNeighbors(edi+2*ecx+1, esi-1)   // down right, up left   
        ChkBetterNeighbors(edi+2*ecx, esi)		// down, up
        ChkBetterNeighbors(edi, esi+2*ecx)		// up, down
        ChkBetterNeighbors(edi+ecx-1, esi+ecx+1) // left, right
        ChkBetterNeighbors(edi+ecx+1, esi+ecx-1) // right, left
		ChkBetterNeighbors(edi+ecx, esi+ecx)	// center, in old and new
// Spatial
        ChkBetterNeighbors(ebx-1, ebx+2*ecx+1)  // up left, down right
        ChkBetterNeighbors(ebx+1, ebx+2*ecx-1)  // up right, down left
        ChkBetterNeighbors(ebx, ebx+2*ecx)	// up, down
//        ChkBetterNeighbors(ebx+ecx-1, ebx+ecx+1) // left, right - no row

// We now know the min and max of the closest together pair of neighbors
		movq	mm0, qword ptr[ebx+ecx]		// our curr pixels vals
		movq	mm1, mm0
		pminub	mm1, mm7					// make it no larger than max neighbor
		pmaxub	mm1, mm6					// and now smaller than min neighbor

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
		add		esi, 8
		add		edi, 8
		add		ebx, 8
		add		edx, 8

		dec		eax
		jnz		LoopQ

		//	do some filtering of last 8 pixels  
		// rewind back to set last column equal to row size - v0.8.2
		mov eax, lastback 
		sub	esi, eax
		sub	edi, eax
		sub ebx, eax
		sub edx, eax
		pcmpeqb	mm5, mm5		// init diff 255's
		pxor	mm6, mm6		// init min to 0
		pcmpeqb	mm7, mm7		// init max 255's

// Search between prev and next frame - Temporal
		// The same row only
        ChkBetterNeighbors(edi+2*ecx, esi)		// down, up
        ChkBetterNeighbors(edi, esi+2*ecx)		// up, down
		ChkBetterNeighbors(edi+ecx, esi+ecx)	// center, in old and new
// Spatial
        ChkBetterNeighbors(ebx, ebx+2*ecx)	// up, down

// We now know the min and max of the closest together pair of neighbors
		movq	mm0, qword ptr[ebx+ecx]		// our curr pixels vals
		movq	mm1, mm0
		pminub	mm1, mm7					// make it no larger than max neighbor
		pmaxub	mm1, mm6					// and now smaller than min neighbor

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
	}

	// adjust for next line
	pSrc  += src_pitch;
	pPrev  += src_pitch;
	pNext  += src_pitch;
	pDest += dst_pitch;
	}
}

//---------------------------------------------------------------------------------------------
// Same function as above except for mode
void DeGrainMedian_YV12mode1norow(BYTE* dstp, int dst_pitch, const BYTE* _pPrev,
			const BYTE* srcp, int src_pitch, const BYTE* _pNext,
			int rowsize, int FldHeight, int Limit)				
{

const BYTE* pPrev = _pPrev;
const BYTE* pNext = _pNext;
const BYTE* pSrc = srcp;

BYTE* pDest = dstp;
int y;
int ct = ((rowsize-9) >> 3);
int lastback = 8 + (ct<<3) + 8 - rowsize; // v0.8.2
__int64	i = Limit;	
__int64	LimitB = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 

	// first and last line are not filtered
	memcpy(pDest, pSrc, rowsize);		// copy first line
	memcpy(pDest+dst_pitch*(FldHeight-1),	pSrc+src_pitch*(FldHeight-1), rowsize);  // and last

	pDest += dst_pitch; // go to second line

	for (y=1; y <= FldHeight-2; y++)	
	{

		__asm
		{
		// Loop general reg usage
		//
		// eax - inner loop counter
		// ebx - src lines above curr 
		// ecx - src pitch
		// edx - dest
		// edi - src pixels, next frame, 1 line up
		// esi - src pixels, prev frame, 1 line up

		mov		ebx, pSrc   
		mov		esi, pPrev
		mov		edi, pNext
		mov		ecx, src_pitch; //src_pitch 
		mov		edx, pDest 

//		pxor	mm4, mm4						// useful constant 000..	


		//	do some filtering of first 8 pixels  

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// Search Spatial
		diagweight8more(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight8more(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
		// make some adjustments to start output on 2nd qword of 2nd line
		add		edi, 8
		add		ebx, 8
		add		esi, 8
		add		edx, 8
		mov	    eax, ct			// loop ct

LoopQ:	

// Search between prev and next frame - Temporal

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// diagweight8(oldp, newp, weight, bound1, bound2, reg1, reg2)	
// oldp is clipped with min(bound1, bound2) and max(bound1, bound2)
// finally weight = |oldp - newp| + 2*|bound1 - bound2|
// Search Spatial
		diagweight8more(mm0, mm1, mm2, [ebx - 1], [ebx + 2*ecx + 1], mm6, mm7) // up left, down right
		diagweight8more(mm0, mm3, mm4, [ebx + 2*ecx -1], [ebx + 1], mm6, mm7)// up right, down left
//	mergeweighted(value1, weight1, value2, weight2) 
// if( weight2[i] <= weight1[i] ) { value1[i] = value2[i]; weight1[i] = weight2[i]; }
// value2 remains unchanged
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
//		diagweight8more(mm0, mm3, mm4, [ebx + ecx + 1], [ebx + ecx - 1], mm6, mm7)// left, right - no row
//		mergeweighted(mm1, mm2, mm3, mm4)
// Search between prev and next frame - Temporal
		diagweight8more(mm0, mm3, mm4, [edi-1], [esi+2*ecx+1], mm6, mm7)// up left, down right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+1], [esi+2*ecx-1], mm6, mm7)// up right, down left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+2*ecx-1], [esi+1], mm6, mm7)// down left, up right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+2*ecx+1], [esi-1], mm6, mm7)// down right, up left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+ecx-1], [esi+ecx+1], mm6, mm7)// left, right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+ecx+1], [esi+ecx-1], mm6, mm7)// right, left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..
		
		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
		add		esi, 8
		add		edi, 8
		add		ebx, 8
		add		edx, 8

		dec		eax
		jnz		LoopQ

		//	do some filtering of last 8 pixels  
		// rewind back to set last column equal to row size - v0.8.2
		mov eax, lastback 
		sub	esi, eax
		sub	edi, eax
		sub ebx, eax
		sub edx, eax
		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

		// Search Spatial
		diagweight8more(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight8more(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8more(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
	}

	// adjust for next line
	pSrc  += src_pitch;
	pPrev  += src_pitch;
	pNext  += src_pitch;
	pDest += dst_pitch;
	}
}

//---------------------------------------------------------------------------------------------
// Same function as above except for mode
void DeGrainMedian_YV12mode2norow(BYTE* dstp, int dst_pitch, const BYTE* _pPrev,
			const BYTE* srcp, int src_pitch, const BYTE* _pNext,
			int rowsize, int FldHeight, int Limit)				
{

const BYTE* pPrev = _pPrev;
const BYTE* pNext = _pNext;
const BYTE* pSrc = srcp;

BYTE* pDest = dstp;
int y;
int ct = ((rowsize-9) >> 3);
int lastback = 8 + (ct<<3) + 8 - rowsize; // v0.8.2
__int64	i = Limit;	
__int64	LimitB = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 

	// first and last line are not filtered
	memcpy(pDest, pSrc, rowsize);		// copy first line
	memcpy(pDest+dst_pitch*(FldHeight-1),	pSrc+src_pitch*(FldHeight-1), rowsize);  // and last

	pDest += dst_pitch; // go to second line

	for (y=1; y <= FldHeight-2; y++)	
	{

		__asm
		{
		// Loop general reg usage
		//
		// eax - inner loop counter
		// ebx - src lines above curr 
		// ecx - src pitch
		// edx - dest
		// edi - src pixels, next frame, 1 line up
		// esi - src pixels, prev frame, 1 line up

		mov		ebx, pSrc   
		mov		esi, pPrev
		mov		edi, pNext
		mov		ecx, src_pitch; //src_pitch 
		mov		edx, pDest 

//		pxor	mm4, mm4						// useful constant 000..	


		//	do some filtering of first 8 pixels  

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// Search Spatial
		diagweight8(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight8(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
		// make some adjustments to start output on 2nd qword of 2nd line
		add		edi, 8
		add		ebx, 8
		add		esi, 8
		add		edx, 8
		mov	    eax, ct			// loop ct

LoopQ:	

// Search between prev and next frame - Temporal

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// diagweight8(oldp, newp, weight, bound1, bound2, reg1, reg2)	
// oldp is clipped with min(bound1, bound2) and max(bound1, bound2)
// finally weight = |oldp - newp| + 2*|bound1 - bound2|
// Search Spatial
		diagweight8(mm0, mm1, mm2, [ebx - 1], [ebx + 2*ecx + 1], mm6, mm7) // up left, down right
		diagweight8(mm0, mm3, mm4, [ebx + 2*ecx -1], [ebx + 1], mm6, mm7)// up right, down left
//	mergeweighted(value1, weight1, value2, weight2) 
// if( weight2[i] <= weight1[i] ) { value1[i] = value2[i]; weight1[i] = weight2[i]; }
// value2 remains unchanged
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
//		diagweight8(mm0, mm3, mm4, [ebx + ecx + 1], [ebx + ecx - 1], mm6, mm7)// left, right - norow
//		mergeweighted(mm1, mm2, mm3, mm4)
// Search between prev and next frame - Temporal
		diagweight8(mm0, mm3, mm4, [edi-1], [esi+2*ecx+1], mm6, mm7)// up left, down right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+1], [esi+2*ecx-1], mm6, mm7)// up right, down left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+2*ecx-1], [esi+1], mm6, mm7)// down left, up right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+2*ecx+1], [esi-1], mm6, mm7)// down right, up left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+ecx-1], [esi+ecx+1], mm6, mm7)// left, right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+ecx+1], [esi+ecx-1], mm6, mm7)// right, left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..
		
		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
		add		esi, 8
		add		edi, 8
		add		ebx, 8
		add		edx, 8

		dec		eax
		jnz		LoopQ

		//	do some filtering of last 8 pixels  
		// rewind back to set last column equal to row size - v0.8.2
		mov eax, lastback 
		sub	esi, eax
		sub	edi, eax
		sub ebx, eax
		sub edx, eax
		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

		// Search Spatial
		diagweight8(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight8(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight8(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
	}

	// adjust for next line
	pSrc  += src_pitch;
	pPrev  += src_pitch;
	pNext  += src_pitch;
	pDest += dst_pitch;
	}
}

//---------------------------------------------------------------------------------------------
// Same function as above except for mode
void DeGrainMedian_YV12mode3norow(BYTE* dstp, int dst_pitch, const BYTE* _pPrev,
			const BYTE* srcp, int src_pitch, const BYTE* _pNext,
			int rowsize, int FldHeight, int Limit)				
{

const BYTE* pPrev = _pPrev;
const BYTE* pNext = _pNext;
const BYTE* pSrc = srcp;

BYTE* pDest = dstp;
int y;
int ct = ((rowsize-9) >> 3);
int lastback = 8 + (ct<<3) + 8 - rowsize; // v0.8.2
__int64	i = Limit;	
__int64	LimitB = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 

	// first and last line are not filtered
	memcpy(pDest, pSrc, rowsize);		// copy first line
	memcpy(pDest+dst_pitch*(FldHeight-1),	pSrc+src_pitch*(FldHeight-1), rowsize);  // and last

	pDest += dst_pitch; // go to second line

	for (y=1; y <= FldHeight-2; y++)	
	{

		__asm
		{
		// Loop general reg usage
		//
		// eax - inner loop counter
		// ebx - src lines above curr 
		// ecx - src pitch
		// edx - dest
		// edi - src pixels, next frame, 1 line up
		// esi - src pixels, prev frame, 1 line up

		mov		ebx, pSrc   
		mov		esi, pPrev
		mov		edi, pNext
		mov		ecx, src_pitch; //src_pitch 
		mov		edx, pDest 

//		pxor	mm4, mm4						// useful constant 000..	


		//	do some filtering of first 8 pixels  

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// Search Spatial
		diagweight7(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight7(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
		// make some adjustments to start output on 2nd qword of 2nd line
		add		edi, 8
		add		ebx, 8
		add		esi, 8
		add		edx, 8
		mov	    eax, ct			// loop ct

LoopQ:	

// Search between prev and next frame - Temporal

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// diagweight7(oldp, newp, weight, bound1, bound2, reg1, reg2)	
// oldp is clipped with min(bound1, bound2) and max(bound1, bound2)
// finally weight = |oldp - newp| + 2*|bound1 - bound2|
// Search Spatial
		diagweight7(mm0, mm1, mm2, [ebx - 1], [ebx + 2*ecx + 1], mm6, mm7) // up left, down right
		diagweight7(mm0, mm3, mm4, [ebx + 2*ecx -1], [ebx + 1], mm6, mm7)// up right, down left
//	mergeweighted(value1, weight1, value2, weight2) 
// if( weight2[i] <= weight1[i] ) { value1[i] = value2[i]; weight1[i] = weight2[i]; }
// value2 remains unchanged
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
//		diagweight7(mm0, mm3, mm4, [ebx + ecx + 1], [ebx + ecx - 1], mm6, mm7)// left, right - no row
//		mergeweighted(mm1, mm2, mm3, mm4)
// Search between prev and next frame - Temporal
		diagweight7(mm0, mm3, mm4, [edi-1], [esi+2*ecx+1], mm6, mm7)// up left, down right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+1], [esi+2*ecx-1], mm6, mm7)// up right, down left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+2*ecx-1], [esi+1], mm6, mm7)// down left, up right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+2*ecx+1], [esi-1], mm6, mm7)// down right, up left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+ecx-1], [esi+ecx+1], mm6, mm7)// left, right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+ecx+1], [esi+ecx-1], mm6, mm7)// right, left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..
		
		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
		add		esi, 8
		add		edi, 8
		add		ebx, 8
		add		edx, 8

		dec		eax
		jnz		LoopQ

		//	do some filtering of last 8 pixels  
		// rewind back to set last column equal to row size - v0.8.2
		mov eax, lastback 
		sub	esi, eax
		sub	edi, eax
		sub ebx, eax
		sub edx, eax
		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

		// Search Spatial
		diagweight7(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight7(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight7(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
	}

	// adjust for next line
	pSrc  += src_pitch;
	pPrev  += src_pitch;
	pNext  += src_pitch;
	pDest += dst_pitch;
	}
}

//---------------------------------------------------------------------------------------------
// Same function as above except for mode
void DeGrainMedian_YV12mode4norow(BYTE* dstp, int dst_pitch, const BYTE* _pPrev,
			const BYTE* srcp, int src_pitch, const BYTE* _pNext,
			int rowsize, int FldHeight, int Limit)				
{

const BYTE* pPrev = _pPrev;
const BYTE* pNext = _pNext;
const BYTE* pSrc = srcp;

BYTE* pDest = dstp;
int y;
int ct = ((rowsize-9) >> 3);
int lastback = 8 + (ct<<3) + 8 - rowsize; // v0.8.2
__int64	i = Limit;	
__int64	LimitB = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 

	// first and last line are not filtered
	memcpy(pDest, pSrc, rowsize);		// copy first line
	memcpy(pDest+dst_pitch*(FldHeight-1),	pSrc+src_pitch*(FldHeight-1), rowsize);  // and last

	pDest += dst_pitch; // go to second line

	for (y=1; y <= FldHeight-2; y++)	
	{

		__asm
		{
		// Loop general reg usage
		//
		// eax - inner loop counter
		// ebx - src lines above curr 
		// ecx - src pitch
		// edx - dest
		// edi - src pixels, next frame, 1 line up
		// esi - src pixels, prev frame, 1 line up

		mov		ebx, pSrc   
		mov		esi, pPrev
		mov		edi, pNext
		mov		ecx, src_pitch; //src_pitch 
		mov		edx, pDest 

//		pxor	mm4, mm4						// useful constant 000..	


		//	do some filtering of first 8 pixels  

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// Search Spatial
		diagweight6(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight6(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
		// make some adjustments to start output on 2nd qword of 2nd line
		add		edi, 8
		add		ebx, 8
		add		esi, 8
		add		edx, 8
		mov	    eax, ct			// loop ct

LoopQ:	

// Search between prev and next frame - Temporal

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// diagweight6(oldp, newp, weight, bound1, bound2, reg1, reg2)	
// oldp is clipped with min(bound1, bound2) and max(bound1, bound2)
// finally weight = |oldp - newp| + 2*|bound1 - bound2|
// Search Spatial
		diagweight6(mm0, mm1, mm2, [ebx - 1], [ebx + 2*ecx + 1], mm6, mm7) // up left, down right
		diagweight6(mm0, mm3, mm4, [ebx + 2*ecx -1], [ebx + 1], mm6, mm7)// up right, down left
//	mergeweighted(value1, weight1, value2, weight2) 
// if( weight2[i] <= weight1[i] ) { value1[i] = value2[i]; weight1[i] = weight2[i]; }
// value2 remains unchanged
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
//		diagweight6(mm0, mm3, mm4, [ebx + ecx + 1], [ebx + ecx - 1], mm6, mm7)// left, right - norow
//		mergeweighted(mm1, mm2, mm3, mm4)
// Search between prev and next frame - Temporal
		diagweight6(mm0, mm3, mm4, [edi-1], [esi+2*ecx+1], mm6, mm7)// up left, down right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+1], [esi+2*ecx-1], mm6, mm7)// up right, down left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+2*ecx-1], [esi+1], mm6, mm7)// down left, up right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+2*ecx+1], [esi-1], mm6, mm7)// down right, up left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+ecx-1], [esi+ecx+1], mm6, mm7)// left, right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+ecx+1], [esi+ecx-1], mm6, mm7)// right, left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..
		
		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
		add		esi, 8
		add		edi, 8
		add		ebx, 8
		add		edx, 8

		dec		eax
		jnz		LoopQ

		//	do some filtering of last 8 pixels  
		// rewind back to set last column equal to row size - v0.8.2
		mov eax, lastback 
		sub	esi, eax
		sub	edi, eax
		sub ebx, eax
		sub edx, eax
		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

		// Search Spatial
		diagweight6(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight6(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight6(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
	}

	// adjust for next line
	pSrc  += src_pitch;
	pPrev  += src_pitch;
	pNext  += src_pitch;
	pDest += dst_pitch;
	}
}

//---------------------------------------------------------------------------------------------
// Same function as above except for mode
void DeGrainMedian_YV12mode5norow(BYTE* dstp, int dst_pitch, const BYTE* _pPrev,
			const BYTE* srcp, int src_pitch, const BYTE* _pNext,
			int rowsize, int FldHeight, int Limit)				
{

const BYTE* pPrev = _pPrev;
const BYTE* pNext = _pNext;
const BYTE* pSrc = srcp;

BYTE* pDest = dstp;
int y;
int ct = ((rowsize-9) >> 3);
int lastback = 8 + (ct<<3) + 8 - rowsize; // v0.8.2
__int64	i = Limit;	
__int64	LimitB = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 

	// first and last line are not filtered
	memcpy(pDest, pSrc, rowsize);		// copy first line
	memcpy(pDest+dst_pitch*(FldHeight-1),	pSrc+src_pitch*(FldHeight-1), rowsize);  // and last

	pDest += dst_pitch; // go to second line

	for (y=1; y <= FldHeight-2; y++)	
	{

		__asm
		{
		// Loop general reg usage
		//
		// eax - inner loop counter
		// ebx - src lines above curr 
		// ecx - src pitch
		// edx - dest
		// edi - src pixels, next frame, 1 line up
		// esi - src pixels, prev frame, 1 line up

		mov		ebx, pSrc   
		mov		esi, pPrev
		mov		edi, pNext
		mov		ecx, src_pitch; //src_pitch 
		mov		edx, pDest 

//		pxor	mm4, mm4						// useful constant 000..	


		//	do some filtering of first 8 pixels  

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// Search Spatial
		diagweight5(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight5(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
		// make some adjustments to start output on 2nd qword of 2nd line
		add		edi, 8
		add		ebx, 8
		add		esi, 8
		add		edx, 8
		mov	    eax, ct			// loop ct

LoopQ:	

// Search between prev and next frame - Temporal

		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

// diagweight5(oldp, newp, weight, bound1, bound2, reg1, reg2)	
// oldp is clipped with min(bound1, bound2) and max(bound1, bound2)
// finally weight = |oldp - newp| + 2*|bound1 - bound2|
// Search Spatial
		diagweight5(mm0, mm1, mm2, [ebx - 1], [ebx + 2*ecx + 1], mm6, mm7) // up left, down right
		diagweight5(mm0, mm3, mm4, [ebx + 2*ecx -1], [ebx + 1], mm6, mm7)// up right, down left
//	mergeweighted(value1, weight1, value2, weight2) 
// if( weight2[i] <= weight1[i] ) { value1[i] = value2[i]; weight1[i] = weight2[i]; }
// value2 remains unchanged
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
//		diagweight5(mm0, mm3, mm4, [ebx + ecx + 1], [ebx + ecx - 1], mm6, mm7)// left, right - no row
//		mergeweighted(mm1, mm2, mm3, mm4)
// Search between prev and next frame - Temporal
		diagweight5(mm0, mm3, mm4, [edi-1], [esi+2*ecx+1], mm6, mm7)// up left, down right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+1], [esi+2*ecx-1], mm6, mm7)// up right, down left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+2*ecx-1], [esi+1], mm6, mm7)// down left, up right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+2*ecx+1], [esi-1], mm6, mm7)// down right, up left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+ecx-1], [esi+ecx+1], mm6, mm7)// left, right
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+ecx+1], [esi+ecx-1], mm6, mm7)// right, left
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..
		
		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
		add		esi, 8
		add		edi, 8
		add		ebx, 8
		add		edx, 8

		dec		eax
		jnz		LoopQ

		//	do some filtering of last 8 pixels  
		// rewind back to set last column equal to row size - v0.8.2
		mov eax, lastback 
		sub	esi, eax
		sub	edi, eax
		sub ebx, eax
		sub edx, eax
		movq	mm0, qword ptr[ebx+ecx]		// load our curr pixels vals

		// Search Spatial
		diagweight5(mm0, mm1, mm2, [ebx + 2*ecx ], [ebx], mm6, mm7)// up, down
// Search between prev and next frame - Temporal
		diagweight5(mm0, mm3, mm4, [edi+2*ecx], [esi], mm6, mm7)//  down, up
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi], [esi+2*ecx], mm6, mm7)// up, down
		mergeweighted(mm1, mm2, mm3, mm4)
		diagweight5(mm0, mm3, mm4, [edi+ecx], [esi+ecx], mm6, mm7)// center, in old and new
		mergeweighted(mm1, mm2, mm3, mm4)

		pxor	mm4, mm4						// useful constant 000..	

		// Limit pixel correction 
		LimitPixelCorrection(LimitB)

		movntq	qword ptr[edx], mm1			// and store it
	
	}

	// adjust for next line
	pSrc  += src_pitch;
	pPrev  += src_pitch;
	pNext  += src_pitch;
	pDest += dst_pitch;
	}
}


